// -*- Mode:C++ -*-
//Header:
//File: mmtoolT.h
//Title: Memory Manager Templates
//Author: NODA Itsuki
//Date: 1998/4/23
//EndHeader:

//ModifyHistory:
// 1998/4/23: Create this file
//EndModifyHistory:

//Memo:
// [Contents]
//	class AMemory<class Content>
//		Arrayed memory
//	class APMemory<class Content>
//		Arrayed pointered memory
//	class AStack<class Content>
//		Stack managed in array
//	class AHeap<class Content>
//		Heap managed in array
//	class SpStack<class Content>
//	class SpStackEntry<class Content>
//		Spagetti Stack
//	class RingBuffer<class Content>
//		Ringed buffer
//	class PRHeap
//		Passive Recycle Heap
//	class SList
//	class SListHeap
//		Sequential List
//EndMemo:

#ifndef _mmtoolT_h_
#define _mmtoolT_h_
////////////////////////////////////////////////////////////////////////

//--------------------------------------------------
// includes
#include <stdlib.h>
#include <iostream.h>
#include "base.h"

//--------------------------------------------------
// constants

#ifndef DEFAULT_ARRAY_SIZE
#define DEFAULT_ARRAY_SIZE	0
#endif
#ifndef DEFAULT_EXPAND_SIZE
#define DEFAULT_EXPAND_SIZE	64
#endif
#ifndef DEFAULT_EXPAND_FACTOR
#define DEFAULT_EXPAND_FACTOR	2
#endif
#ifndef DEFAULT_MEMORY_NAME
#define DEFAULT_MEMORY_NAME	"Memory"
#endif

#ifndef VERBOSE_EXPAND_P
#define VERBOSE_EXPAND_P	false
#endif


//======================================================================
// AMemory<Content>
//	arrayed memory 
template <class _Contents_>
class AMemory {
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef AMemory<_Contents_> amemory_t ; 
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Member Variables

      cont_t* table ;
      UInt _top ;
      UInt size ;
      UInt expand_factor ;
      char* name ;

      //--------------------------------------------------
      // Constructor

      inline AMemory<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				 UInt sz = DEFAULT_ARRAY_SIZE,
				 UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm,sz,ef) ;
      } ;

      inline bool init(char* nm = DEFAULT_MEMORY_NAME, 
		       UInt sz = DEFAULT_MEMORY_NAME, 
		       UInt ef = DEFAULT_EXPAND_FACTOR) {
	 name = nm ;
	 size = 0 ;
	 expand_factor = ef ;
	 _top = 0 ;
	 table = NULLPTR ;
	 return expand(sz,false) ;
      } ;
      
      //--------------------------------------------------
      // Expand

      bool expand(bool verbosep = VERBOSE_EXPAND_P) {
	 if(size>0) return expand(size * expand_factor, verbosep) ;
	 else return expand(DEFAULT_EXPAND_SIZE, verbosep) ;
      } ;

      bool expand(UInt newsize, bool verbosep) {
	 if(verbosep) {
	    cerr << "Expanding " << *this << ": " << size << "->" << newsize 
	       << " ... " ;
	    cerr.flush() ;
	 } 
	 if(newsize == 0 || newsize < size) return false ;
	 cont_t* oldtable = table ;
	 table = new cont_t[newsize] ;
	 if(table == NULLPTR) {
	    cerr << endl ;
	    cerr << "Warning: can't expand the size of " ;
	    describe(cerr,false) ;
	    cerr << "." << endl ;
	    return false ;
	 } 
	 for(UInt i = 0 ; i < size ; i++) {
	    table[i] = oldtable[i] ;
	 } 
	 size = newsize ;
	 if(verbosep) {
	    cerr << "Expanding is done." << endl ;
	 }
	 delete oldtable ;
	 return true ;
      } ;

      //--------------------------------------------------
      // describe

      void describe(ostream& ostr, bool detailp = true) const {
	 ostr << "#" << name << "[" << this 
	      << ":size=" << size 
	      << ",top=" << _top 
	      << ",ef=" << expand_factor 
	      << ",table=" << table 
	      << "]" ;
	 if(detailp) {
	    ostr << endl ;
	    for(UInt i = 0 ; i < _top ; i++) {
	       ostr << "\t" << table[i] << endl ;
	    }
	 }
      } ;

      friend ostream& operator<< (ostream& o, const amemory_t& am) {
	 am.describe(o, false) ;
	 return o ;
      } ;
} ;
      

//======================================================================
// APMemory
//    arrayed pointered memory
template <class _Contents_>
class APMemory : public AMemory<_Contents_ *> {
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef APMemory<_Contents_> apmemory_t ; 
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Member Variables
      
      //--------------------------------------------------
      // Constructor

      inline APMemory<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				  UInt sz = DEFAULT_ARRAY_SIZE,
				  UInt ef = DEFAULT_EXPAND_FACTOR) 
      {
	 init(nm,sz,ef) ;
      } ;

      bool init(char *nm, UInt sz, UInt ef) {
	 ((amemory_t*)this)->init(nm, 0, ef) ;
	 return expand(sz,false) ; 
      } ;

      //--------------------------------------------------
      // Expand
      
      bool expand(bool verbosep = VERBOSE_EXPAND_P) {
	 if(size>0) return expand(size * expand_factor, verbosep) ;
	 else return expand(DEFAULT_EXPAND_SIZE, verbosep) ;
      } ;

      bool expand(UInt newsize, bool verbosep) {
	 UInt oldsize = size ;
	 UInt sizedif = newsize - oldsize ;

	 ((amemory_t*)this)->expand(newsize, verbosep) ;

	 cont_t* newmem = new cont_t[sizedif] ;
	 if(newmem == NULLPTR) {
	    cerr << endl ;
	    cerr << "Warning: can't allocate new memory for " ;
	    describe(cerr,false) ;
	    cerr << "." << endl ;
	    return false ;
	 } 
	 for(UInt i = 0 ; i < sizedif ; i++) {
	    table[i+oldsize] = &(newmem[i]) ;
	 }
	 if(verbosep) {
	    cerr << "  Allocating is done." << endl ;
	 }
	 return true ;
      } ;

      //--------------------------------------------------
      // describe

      void describe(ostream& ostr, bool detailp = true) const {
	 ((amemory_t*)this)->describe(ostr, false) ;
	 if(detailp) {
	    ostr << endl ;
	    for(UInt i = 0 ; i < _top ; i++) {
	       ostr << "\t" << *(table[i]) << endl ;
	    }
	 }
      } ;

      friend ostream& operator<< (ostream& o, const apmemory_t& am) {
	 am.describe(o, false) ;
	 return o ;
      } ;
} ;


//======================================================================
// AStack<Content>
//
template <class _Contents_>
class AStack : public AMemory<_Contents_>{
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef AStack<_Contents_> astack_t ; 

      //--------------------------------------------------
      // Constructor

      inline AStack<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				UInt sz = DEFAULT_ARRAY_SIZE,
				UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm,sz,ef) ;
      } ;

      //--------------------------------------------------
      // Push/Pop

      astack_t* push(const cont_t& val) {
	 if(_top >= size) expand() ;
	 table[_top] = val ;
	 _top++ ;
	 return this ;
      } ;

      cont_t pop() {
	 if(_top <= 0) {
	    cerr << "Error: pop operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 _top-- ;
	 return table[_top] ;
      } ;
      
      cont_t popN(UInt c) {
	 if(_top < c) {
	    cerr << "Error: pop operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 _top -= c ;
	 return table[_top] ;
      } ;

      bool removeN(UInt n) {
	 if(_top <= n) {
	    return false ;
	 }
	 for(UInt i = 0 ; i < n ; i++) {
	    table[_top-i-1] = table[_top-i] ;
	 } 
	 _top-- ;
	 return true ;
      } ;
	 

      //--------------------------------------------------
      // access to top stack

      cont_t& top() const {
	 if(_top <= 0) {
	    cerr << "Error: top operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 return table[_top-1] ;
      } ;

      cont_t& topN(UInt c) const {
	 if(_top < c) {
	    cerr << "Error: topN operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 return table[_top-c-1] ;
      } ;

      inline UInt toppos() const { return _top ; } ;
      inline UInt setpos(UInt newtop) { return _top = newtop ; } ;
      inline UInt reset() { return setpos(0) ; } ;

      inline bool bottomp() const { return _top == 0 ; } ;

      inline void describe(ostream& o, bool detailp = true) const {
	 ((amemory_t*)this)->describe(o, false) ;
	 if(detailp) {
	    o << endl ;
	    for(UInt i = 0 ; i < _top ; i++) {
	       o << "\t" << topN(i) << endl ;
	    }
	    o << "\t___bottom___" << endl ;
	 }
      } ;

      friend ostream& operator<< (ostream& o, const astack_t& st) {
	 st.describe(o, false) ;
	 return o ;
      } ;
} ;
	  

//======================================================================
// AHeap<Content>
//
template <class _Contents_>
class AHeap : public APMemory<_Contents_> {
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef AHeap<_Contents_> aheap_t ; 

      //--------------------------------------------------
      // Constructor

      inline AHeap<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
			       UInt sz = DEFAULT_ARRAY_SIZE,
			       UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm,sz,ef) ;
      } ;

      //--------------------------------------------------
      // get new entity

      cont_t* get() {
	 if(_top >= size) expand() ;
	 cont_t* r = table[_top] ;
	 r->init() ;
	 _top++ ;
	 return r ;
      } ;

      //--------------------------------------------------
      // access to top stack

      inline UInt toppos() { return _top ; } ;
     
      inline bool rewind(const UInt pos) { _top = pos ; return true ; } ;

      inline bool reset() { return rewind(0) ; } ;

};


//======================================================================
// SpStack<Content>
//	Spagetti Stack

template <class _Contents_>
class SpStackEntry {
   public:
      //--------------------------------------------------
      // LocalTypeDef
      typedef SpStackEntry<_Contents_> ss_entry_t ;
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Member Variables

      cont_t entity ;
      ss_entry_t* next ;

      //--------------------------------------------------
      // Constructor
      
      inline SpStackEntry<_Contents_>(ss_entry_t* nx = NULLPTR) {
	 next = nx ; 
      } ;
      inline SpStackEntry<_Contents_>(cont_t& ent, ss_entry_t nx = NULLPTR) {
	 entity = ent ;
	 next = nx ;
      } ;

      //--------------------------------------------------
      // Initializer
      bool init() {
	 next = NULLPTR ;
      } ;	 

      //--------------------------------------------------
      // Describe

      void describe(ostream& o, bool detailp = true) const {
	 if(detailp) {
	    o << "#SpStackEntry[" << this << "]" << endl ; 
	    for(const ss_entry_t* p = this ; p != NULLPTR ; p = p->next ) {
	       o << "\t" << entity << endl ;
	    }
	 } else {
	    o << "#SpStackEntry[" << entity << "]" ;
	 }
      } ;

      friend ostream& operator<< (ostream& o, const ss_entry_t& ss) {
	 ss.describe(o, false) ;
	 return o ;
      } ;
} ;

template <class _Contents_>
class SpStack : public AHeap< SpStackEntry<_Contents_> >{
   public:
      typedef SpStack<_Contents_> spstack_t ;
      typedef SpStackEntry<_Contents_> ss_entry_t ;
      typedef _Contents_ cont_t ;
      class rewindPos {
	 public:
	    ss_entry_t* stacktop ;
	    UInt	heaptop ;
	    
	    inline rewindPos() { 
	       stacktop = NULLPTR ; heaptop = 0 ;
	    } ;
	    inline rewindPos(const rewindPos& rpos) {
	       stacktop = rpos.stacktop ; 
	       heaptop = rpos.heaptop ;
	    } ;
	    inline rewindPos(ss_entry_t* stop, UInt htop) {
	       stacktop = stop ;
	       heaptop = htop ;
	    } ;
	    inline rewindPos operator=(const rewindPos& rpos) {
	       stacktop = rpos.stacktop ; 
	       heaptop = rpos.heaptop ;
	       return *this;
	    } ;
	    
	    inline friend ostream& operator<<(ostream& o, 
					      const rewindPos& rpos) 
	    { 
	       o << "#rewindPos[stacktop=" << rpos.stacktop ;
	       o << ", heaptop=" << rpos.heaptop ; 
	       return o ;
	    } ;
      } ;

      //--------------------------------------------------
      // Member Variables

      ss_entry_t* stackTop ;

      //--------------------------------------------------
      // Constructor
      
      inline SpStack<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				 UInt sz = DEFAULT_ARRAY_SIZE,
				 UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm, sz, ef) ;
      } ;

      bool init(char *nm, UInt sz, UInt ef) {
	 ((aheap_t*)this)->init(nm,sz,ef) ;
	 stackTop = NULLPTR ;
      } ;
	    
      //--------------------------------------------------
      // Push/Pop

      cont_t* pushget() {
	 ss_entry_t* newtop = get() ;
	 newtop->next = stackTop ;
	 stackTop = newtop ;
	 return &(newtop->entity) ;
      } ;

      ss_entry_t* push(const cont_t& ent) {
	 ss_entry_t* newtop = get() ;
	 newtop->entity = ent ;
	 newtop->next = stackTop ;
	 stackTop = newtop ;
	 return newtop ;
      } ;

      ss_entry_t* _pop() {
	 ss_entry_t* oldtop = stackTop ;
	 stackTop = oldtop->next ;
	 return oldtop ;
      } ;

      inline cont_t& pop() {
	 return _pop()->entity ;
      } ;

      inline cont_t& top() const {
	 return stackTop->entity ;
      } ;
      inline cont_t* toppointer() const {
	 return &(stackTop->entity) ;
      } ;

      cont_t& topN(UInt c) const {
	 ss_entry_t* ent = stackTop ;
	 for(UInt i = 0 ; i < c ; i++) {
	    if(ent == NULLPTR) break ;
	    ent = ent->next ;
	 }
	 if(ent == NULLPTR) {
	    cerr << "Error: topN operation hits stack bottom (" << *this
		 << ")." << endl ;
	    exit(1) ;
	 }
	 return ent->entity ;
      } ;

      inline bool bottomp() const {
	 return isNullPtr(stackTop) ;
      } ;

      //--------------------------------------------------
      // rewind

      inline rewindPos rewindpos() {
	 return rewindPos(stackTop, _top) ;
      } ;

      inline bool rewind(rewindPos& rpos) {
	 return rewind(rpos.stacktop, rpos.heaptop) ;
      } ;
	 
      inline bool rewind(ss_entry_t* stacktop, UInt heaptop) { 
	 stackTop = stacktop ;
	 _top = heaptop ; 
	 return true ;
      } ;
      

      inline void setStackTop(ss_entry_t* newtop) {
	 stackTop = newtop ;
      } ;

      inline bool reset() {
	 return rewind(NULLPTR,0) ;
      } ;



      //--------------------------------------------------
      // describe

      void describe(ostream& ostr, bool detailp = true) const {
	 ((aheap_t*)this)->describe(ostr, false) ;
	 if(detailp) {
	    ostr << endl ;
	    for(ss_entry_t* top = stackTop ; top != NULLPTR ; 
					     top = top->next){
	       ostr << "\t" << top->entity << endl ;
	    }
	    ostr << "\t___bottom___" << endl ;
	 } else {
	    ostr << "<" << stackTop << ">" ;
	 }
      } ;

      friend ostream& operator<< (ostream& o, const spstack_t& sps) {
	 sps.describe(o, false) ;
	 return o ;
      } ;
} ;



//======================================================================
// RingBuffer<Content>
//	Ring buffer

template <class _Contents_>
class RingBuffer {
   public:
      //--------------------------------------------------
      // LocalTypeDef

      typedef RingBuffer<_Contents_> ringbuffer_t ;
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Member Variables

      cont_t* table ;
      UInt _top ;
      UInt _bottom ;
      UInt size ;
      UInt expand_factor ;
      char* name ;

      //--------------------------------------------------
      // Constructor

      inline RingBuffer<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				    UInt sz = DEFAULT_ARRAY_SIZE,
				    UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm, sz, ef) ;
      } ;

      bool init(char *nm, UInt sz, UInt ef) {
	 name = nm ;
	 size = 0 ;
	 expand_factor = ef ;
	 _top = 0 ;
	 _bottom = 0 ;
	 expand(sz,false) ;
      } ;
	 
      //--------------------------------------------------
      // Expand

      bool expand(bool verbosep = VERBOSE_EXPAND_P) {
	 if(size>0) return expand(size * expand_factor, verbosep) ;
	 else return expand(DEFAULT_EXPAND_SIZE, verbosep) ;
      } ;

      bool expand(UInt newsize, bool verbosep) {
	 if(verbosep) {
	    cerr << "Expanding " << *this << ": " << size << "->" << newsize 
	       << " ... " ;
	    cerr.flush() ;
	 } 

	 if(newsize == 0 || newsize < size) return false ;
	 cont_t* oldtable = table ;
	 UInt oldsize = size ;
	 table = new cont_t[newsize] ;
	 if(table == NULLPTR) {
	    cerr << endl ;
	    cerr << "Warning: can't expand the size of " ;
	    describe(cerr,false) ;
	    cerr << "." << endl ;
	    return false ;
	 } 

	 for(UInt i = _bottom ; i < _top ; i++) {
	    table[i % newsize] = oldtable[i % oldsize] ;
	 }
	 size = newsize ;
	 reduce_counter() ;
	 
	 if(verbosep) {
	    cerr << "Expanding is done." << endl ;
	 }

	 delete oldtable ;
	 return true ;
      } ;

      inline void reduce_counter() {
	 _top = _top % size ;
	 _bottom = _bottom % size ;
      } ;

      //--------------------------------------------------
      // access to pointer

      inline cont_t* nth_p(UInt n) const {
	 return &(table[n % size]) ;
      } ;

      cont_t* get_p() {
	 if(_top > _bottom) {
	    UInt b = _bottom ;
	    _bottom++ ;
	    if(_bottom > 2 * size) reduce_counter() ;
	    return nth_p(b) ;
	 } else {
	    return NULLPTR ;
	 }
      } ;

      inline bool put_p(cont_t* elm) {
	 return put(*elm) ;
      } ;

      //--------------------------------------------------
      // access to contents

      inline cont_t& nth(const UInt n) const {
	 return *nth_p(n) ;
      } ;

      inline cont_t& get() { /* fragle */
	 return *get_p() ;
      } ;
      inline cont_t& get(cont_t& failval) {
	 cont_t *r = get_p() ;
	 if(r == NULLPTR) return failval ;
	 return *r ;
      } ;
      
      inline bool put(const cont_t& elm, bool verbosep = VERBOSE_EXPAND_P) {
	 if(_top >= _bottom + size) expand(verbosep) ;
	 
	 nth(_top) = elm ;
	 _top++ ; 
	 return true ;
      } ;

      //--------------------------------------------------
      // describe

      void describe(ostream& ostr, bool detailp = true) const {
	 ostr << "#" << name << "[" << this 
	      << ":size=" << size 
	      << ",top/bottom=" << _top << "/" << _bottom
	      << ",ef=" << expand_factor 
	      << ",table=" << table 
	      << "]" ;
	 if(detailp) {
	    ostr << endl ;
	    for(UInt i = _bottom ; i < _top ; i++) {
	       ostr << "\t" << table[i % size] << endl ;
	    }
	 }
      } ;

      friend ostream& operator<< (ostream& o, const ringbuffer_t& rb) {
	 rb.describe(o, false) ;
	 return o ;
      } ;
} ;      

      
      

//======================================================================
// PRHeap
//	Passive Recycle Heap

#define RPH_RINGBUFFERSUFFIX	".RingBuffer"
#define RPH_TABLESUFFIX		".Table"

template <class _Contents_>
class PRHeap {
   public:
      //--------------------------------------------------
      // LocalTypeDef
      typedef PRHeap<_Contents_> prheap_t ;
      typedef _Contents_ cont_t ;
      typedef RingBuffer<_Contents_*> ringbuffer_t ;
      typedef AHeap<_Contents_> aheap_t ;

      //--------------------------------------------------
      // Member Variables
      ringbuffer_t ringBuffer ;
      aheap_t table ;
      char* name ;

      //--------------------------------------------------
      // Constructor
      inline PRHeap<_Contents_>(char *nm = DEFAULT_MEMORY_NAME,
				UInt sz = DEFAULT_ARRAY_SIZE,
				UInt ef = DEFAULT_EXPAND_FACTOR)
      {
	 init(nm, sz, ef) ;
      } ;

      bool init(char* nm, UInt sz, UInt ef){
	 name = nm ;
	 String rb_name = String(nm) + RPH_RINGBUFFERSUFFIX ;
	 String tbl_name = String(nm) + RPH_TABLESUFFIX ;
	 ringBuffer.init((char*)rb_name, sz, ef) ;
	 table.init((char*)tbl_name,sz,ef) ;
	 for(UInt i = 0 ; i < sz ; i++) {
	    put(table.get()) ;
	 }
	 return true ;
      } ;

      //--------------------------------------------------
      // expand
      
      bool expand(bool verbosep = false) {
	 UInt oldsize = table.size ;
	 table.expand(verbosep) ;
	 UInt newsize = table.size ;
	 for(UInt i = oldsize ; i < newsize ; i++) {
	    bool r = put(table.get(),verbosep) ;
	 } ;
	 return true ;
      }

      //--------------------------------------------------
      // get new entity
      
      cont_t* get() {
	 cont_t* r = ringBuffer.get(NULLPTR) ;
	 if(r == NULLPTR) {
	    expand() ;
	    r = ringBuffer.get(NULLPTR) ;
	 }
	 return r ;
      } ;

      //--------------------------------------------------
      // put old entity
	 
      inline bool put(cont_t* elm, bool verbosep = false) {
	 return ringBuffer.put(elm, verbosep) ;
      } ;

      //--------------------------------------------------
      // describe

      void describe(ostream& ostr, bool detailp = true) const {
	 ostr << "#" << name << "[" << this
	      << ":total=" << table.size 
	      << ",ring=" ;
	 ringBuffer.describe(ostr,detailp) ;
	 ostr << "]" ;
	 if(detailp) ostr << endl ;
      }

      friend ostream& operator<< (ostream& o, const prheap_t& ph) {
	 ph.describe(o, false) ;
	 return o ;
      } ;
} ;
	 
//======================================================================
// SList<Content>
//
template <class _Contents_>
class SList {
   public:
      //--------------------------------------------------
      // LocalTypeDef
      typedef SList<_Contents_> slist_t ; 
      typedef _Contents_ cont_t ;

      cont_t element ;
      slist_t* next ;

      inline SList<_Contents_>() {
	 next = NULLPTR ;
      } ;
      inline SList<_Contents_>(const cont_t e) {
	 element = e ;
	 next = NULLPTR ;
      } ;
      inline SList<_Contents_>(const cont_t e, slist_t* n) {
	 element = e ;
	 next = n ;
      } ;

      inline void init() {
	 next = NULLPTR ;
      } ;

      inline bool endp() const { return this == NULLPTR ; } ;

      inline cont_t nth(const UInt n) const {
	 slist_t* p = (slist_t*)this ;
	 for(UInt i = 0 ; i < n ; i++) 
	    if(p != NULLPTR) 
	       p = p->next ;
	 return p->element ;
      } ;
      
      inline UInt length() const {
	 UInt l = 0 ;
	 for(slist_t* p = (slist_t*)this ; !isNullPtr(p) ; p = p->next) l++ ;
	 return l ;
      } ;

      inline friend ostream& operator<< (ostream& o, const slist_t& sl) {
	 o << &sl ;
	 return o ;
      } ;
      inline friend ostream& operator<< (ostream& o, const slist_t* sl) {
	 o << "#SeqList[" ;
	 for(slist_t* p = (slist_t*)sl ; !(p->endp()) ; p = p->next ) 
	    o << p->element ;
	 o << "]" ;
	 return o ;
      } ;
	       
} ;

//======================================================================
// SListHeap<Content>
//
template <class _Contents_>
class SListHeap : public AHeap< SList<_Contents_> > {
   public:	
      //--------------------------------------------------
      // LocalTypeDef
      typedef SList<_Contents_>	slist_t ;
      typedef SListHeap<_Contents_> slheap_t ; 
      typedef _Contents_ cont_t ;

      inline slist_t* push(cont_t el, slist_t* nx) {
	 slist_t* newlist = get() ;
	 newlist->element = el ;
	 newlist->next = nx ;
	 return newlist ;
      } ;

} ;

////////////////////////////////////////////////////////////////////////
#endif


